/****************************************************************************
 * SaneScan 0.52 -- plug-in for the GIMP image manipulation program
 * A scanner plug-in based on the SANE library.
 * Authors:
 *   Andreas Beck (becka@sunserver1.rz.uni-duesseldorf.de) - Original author
 *   Tristan Tarrant (tristan@galactica.it) - Current maintainer
 *   David Mosberger-Tang (davidm@azstarnet.com)
 *
 * You can contact the original "The GIMP" authors at gimp@xcf.berkeley.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "gtk.h"
#ifdef GIMP_PLUGIN
#include "gimp.h"
#endif
#include "sane.h"
#include "saneopts.h"
#include "progress.h"

/* Declare local functions. */
static void error_dialog(gchar *error);

/* Maximum number of controls in device window */
#define MAXdlg	128

/* Information about preview area */
struct PreviewT {
    gint PhysWidth, PhysHeight;
    gint PixelWidth, PixelHeight;
    double Ratio;
    gint PhysX1, PhysY1;
    gint PhysX2, PhysY2;
    gint PixelX1, PixelY1;
    gint PixelX2, PixelY2;
} Preview;

/* Menu description data */
typedef struct MenuItem
{
    gchar *label;
    gchar  accelerator_key;
    gint   accelerator_mods;
    GtkCallback callback;
    gpointer user_data;
    struct MenuItem *subitems;
    GtkWidget *widget;
} MenuItem;

/* Control description */
typedef struct DlgElement {
    GtkWidget *widget, *label;
    GtkData *data;
    MenuItem *menu;
} DlgElement;

struct ScanWinT {
    GtkWidget *shell;		
    GtkWidget *hruler;
    GtkWidget *vruler;
    GtkWidget *preview;
    GtkWidget *info_label;
    GdkGC *gc;

    GtkDataAdjustment *hruledata;	/*  horizontal data information	*/
    GtkDataAdjustment *vruledata;	/*  vertical data information	*/
} ScanWin;

struct SelectionT {
    gint x1, y1, x2, y2;
    gint active;
} Selection, OldSelection;

struct DriverInfoT {
    gint dpi, x1, y1, x2, y2;
} DriverInfo;

gint LeftButtonPressed;

GtkWidget *dialog;
GdkColor RGBPalette[256];
DlgElement elements[MAXdlg];
const SANE_Device **devlist;
static SANE_Handle scan_handle;
gint seldev;		/* The selected device */
gint ndevs;		/* The number of available devices */
#ifdef GIMP_PLUGIN
Image output;
#endif

void PhysToLog(gint xx, gint yy, gint *xret, gint *yret)
{
    gfloat a, b, c, d;
    
    a = (Preview.PixelX2-Preview.PixelX1)/(Preview.PhysX2-Preview.PhysX1);
    b = Preview.PixelX1-a*Preview.PhysX1;
    *xret = (xx-b)/a;
    c = (Preview.PixelX2-Preview.PixelX1)/(Preview.PhysX2-Preview.PhysX1);
    d = Preview.PixelX1-a*Preview.PhysX1;
    *yret = (yy-d)/c;
}

void LogToPhys(gint xx, gint yy, gint *xret, gint *yret)
{
    gfloat a, b, c, d;
    
    a = (Preview.PixelX2-Preview.PixelX1)/(Preview.PhysX2-Preview.PhysX1);
    b = Preview.PixelX1-a*Preview.PhysX1;
    *xret = a*xx+b;
    c = (Preview.PixelX2-Preview.PixelX1)/(Preview.PhysX2-Preview.PhysX1);
    d = Preview.PixelX1-a*Preview.PhysX1;
    *yret = c*yy+d;
}

/*
 * Create a Gtk menu widget from a MenuItem array
 */

GtkWidget *build_menu (MenuItem *items, GtkAcceleratorTable *table)
{
    GtkWidget *menu;
    GtkWidget *menu_item;
    GtkData *data;
    
    menu = gtk_menu_new ();
    
    while (items->label) {
        if (items->label[0] == '-')
            menu_item = gtk_menu_item_new ();
        else {
            menu_item = gtk_menu_item_new_with_label (items->label);
            if (items->accelerator_key && table)
                gtk_widget_install_accelerator (menu_item,
                                                table,
                                                items->accelerator_key,
                                                items->accelerator_mods);
        }
        
        gtk_container_add (menu, menu_item);
        
        if (items->callback) {
            data = gtk_menu_item_get_state (menu_item);
            gtk_callback_add (data, items->callback, items->user_data);
        }
	
        if (items->subitems)
	    gtk_menu_item_set_submenu (menu_item, build_menu (items->subitems, table));
        
        gtk_widget_show (menu_item);
        items->widget = menu_item;
        
        items++;
    }
    
    return menu;
}

int find_menu_option(MenuItem *menu, char *option)
{
    int i;

    i = 0;
    while(strcmp(option,menu[i].label))
	++i;
    return i;
}

/*
 * Update the info line with the latest size information
 */


static void update_param(void)
{ 
    gchar buf[200];
    SANE_Parameters sp;
    
    if (sane_get_parameters(scan_handle,&sp)==SANE_STATUS_GOOD)
	sprintf(buf,"%d x %d : %d KB",sp.pixels_per_line,
		sp.lines,(sp.bytes_per_line*sp.lines)>>10);
    else
	sprintf(buf,"Invalid parameters.");
    if(ScanWin.info_label)
	gtk_label_set(ScanWin.info_label,buf);
}

/*
 * Show/Hide a widget
 */

static void showhide(gint item,gint onoff)
{ 
    if (onoff) {
	if(GTK_WIDGET_VISIBLE (elements[item].widget)) {
	    gtk_widget_hide(elements[item].widget);
	    if(elements[item].label)
		gtk_widget_hide(elements[item].label);
	}
	
    } else {
	if(!GTK_WIDGET_VISIBLE (elements[item].widget)) {
	    gtk_widget_show(elements[item].widget);
	    if(elements[item].label)
		gtk_widget_show(elements[item].label);
	}
    }
}

/* 
 * Update the dialog's widgets with the latest settings
 */

static void update_complete_dialog(void)
{ 
    gint x;
    const SANE_Option_Descriptor *opt;

    for(x=1;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++) {
	switch(opt->type) {
	case SANE_TYPE_BOOL:
	    if (opt->size != sizeof(SANE_Word)) {
		/* Boolean vectors not supported */
		break;
	    }
	    { 
		SANE_Word tog;

		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		gtk_check_button_set_state(elements[x].widget,tog);
		showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		break;
	    } 
	case SANE_TYPE_INT:
	    if (opt->size != sizeof(SANE_Word)) {
		/* Integer vectors not supported yet */
		break;
	    }
	    switch(opt->constraint_type) {
	    case SANE_CONSTRAINT_RANGE: 
		{
		    SANE_Word tog;
		    GtkDataAdjustment *adjdata;

		    tog = 0;
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		    adjdata = (GtkDataAdjustment *) elements[x].data;
		    adjdata->value = (double) (tog) * adjdata->step_increment;
		    gtk_data_notify (elements[x].data);
		    showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		    break;
		}
	    case SANE_CONSTRAINT_WORD_LIST: 
		{
		    SANE_Word val;
		    gchar buf[512];
		    gint i;
		    
		    val = 0;
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&val,NULL);
		    sprintf(buf,"%d",val);
		    i = find_menu_option(elements[x].menu,buf);
		    gtk_option_menu_set_history(elements[x].widget,i);
		    break;
		}
	    default:
		break;
	    }
	    break;
	case SANE_TYPE_FIXED:
	    if (opt->size != sizeof(SANE_Word)) {
		/* Fixed vectors not supported yet */
		break;
	    }
	    switch(opt->constraint_type) {
	    case SANE_CONSTRAINT_RANGE:
		{ 
		    SANE_Word tog;
		    GtkDataAdjustment *adjdata;
		    
		    tog = 0;
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&tog,NULL);
		    tog>>=SANE_FIXED_SCALE_SHIFT;
		    adjdata = (GtkDataAdjustment *) elements[x].data;
		    adjdata->value = (double) (tog) * adjdata->step_increment;
		    gtk_data_notify (elements[x].data);
		    showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		    break;
		}
	    case SANE_CONSTRAINT_WORD_LIST: 
		{
		    SANE_Word val;
		    gchar buf[512];
		    gint i;
		    
		    val = 0;
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&val,NULL);
		    sprintf(buf,"%d",val);
		    i = find_menu_option(elements[x].menu,buf);
		    gtk_option_menu_set_history(elements[x].widget,i);
		    break;
		}
	    default:
		break;
	    } 
	    break;
	case SANE_TYPE_STRING:
	    {
		gchar buf[512];
		gint i;

		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
		switch(opt->constraint_type) {
		case SANE_CONSTRAINT_STRING_LIST:
		    i = find_menu_option(elements[x].menu,buf);
		    gtk_option_menu_set_history(elements[x].widget,i);
		    break;
		case SANE_CONSTRAINT_NONE:
		    gtk_text_entry_set_text(elements[x].widget,buf);
		    break;
		default:
		    break;
		}
		showhide(x,!SANE_OPTION_IS_ACTIVE(opt->cap));
		break;
	    } 
	default:
	    break;
	}
    }
}

void quit_xscan()
{
#ifdef GIMP_PLUGIN
    if(output)
	gimp_free_image(output);
    gimp_noexit_quit();
#endif
    if(scan_handle)
	sane_close(scan_handle);
    sane_exit();
    gtk_exit(0);
}


/*
 * This function is invoked whenever a cancel button is pressed
 */

void quit_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
    quit_xscan();
}

static void error_dialog(gchar *error)
{
    GtkWidget *main_vbox, *hbox, *label;
    GtkWidget *button, *frame, *err_dlg;

    err_dlg = gtk_window_new("Error", GTK_WINDOW_TOPLEVEL);

    /* Create the main vbox */
    main_vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(main_vbox, 0);
    gtk_container_add(err_dlg, main_vbox);
    gtk_widget_show(main_vbox);

    /* The top row with the message */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    /* The message */
    frame = gtk_frame_new(NULL);
    label = gtk_label_new(error);
    gtk_box_pack(hbox, frame, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add(frame,label);
    gtk_widget_show(label);
    gtk_widget_show(frame);

    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    /* The Cancel button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), quit_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    gtk_widget_show(err_dlg);
}

/*
 * This is invoked when the scanning is cancelled
 */

void cancel_scan_callback(GtkWidget *widget,
			  gpointer   client_data,
			  gpointer   call_data)
{
    sane_cancel(scan_handle);
}


/*
 * Invoked when the preview button is pressed
 */

void scan_preview(GtkWidget *widget,
		  gpointer   client_data,
		  gpointer   call_data)
{

}

/*
 * Invoked when the zoom button is pressed
 */

void zoom_in_preview(GtkWidget *widget,
		     gpointer   client_data,
		     gpointer   call_data)
{

    ScanWin.hruledata->lower = Selection.x1;
    ScanWin.hruledata->upper = Selection.x2;
    ScanWin.vruledata->lower = Selection.y1;
    ScanWin.vruledata->upper = Selection.y2;

    gtk_data_notify ((GtkData*) ScanWin.hruledata);
    gtk_data_notify ((GtkData*) ScanWin.vruledata);

}

void zoom_out_preview(GtkWidget *widget,
		      gpointer   client_data,
		      gpointer   call_data)
{

    ScanWin.hruledata->lower = 0;
    ScanWin.hruledata->upper = Preview.PhysWidth;
    ScanWin.vruledata->lower = 0;
    ScanWin.vruledata->upper = Preview.PhysHeight;

    gtk_data_notify ((GtkData*) ScanWin.hruledata);
    gtk_data_notify ((GtkData*) ScanWin.vruledata);

}


/*
 * Invoked when the scan button is pressed
 */

void scan_dialog(GtkWidget *widget,
		 gpointer   client_data,
		 gpointer   call_data)
{
    SANE_Parameters parm;
    gchar *d;
    gint len,cnt,status,total;
    ProgressP scanprogress;
    FILE *fout;

#ifdef GIMP_PLUGIN
    output=NULL;
#endif
    /* Initialize device */
    status=sane_start(scan_handle);
    /* Is everything ok ? */
    if (status!=SANE_STATUS_GOOD) {
	error_dialog("Sane_start failed !");
	return;
    }
    sane_get_parameters(scan_handle,&parm);
#ifdef GIMP_PLUGIN
    if (parm.first_frame)	/* Open a new image */
	output = gimp_new_image (0,parm.pixels_per_line,parm.lines,
				 (parm.format==SANE_FRAME_GRAY) ? GRAY_IMAGE : RGB_IMAGE );
    d = gimp_image_data (output);
    
#else
    fout = fopen("sane_output.ppm","w");
    d = (guchar *)malloc(5000);
    switch (parm.format) {
    case SANE_FRAME_RGB:
	fprintf(fout,"P6\n# SANE data follows\n%d %d\n255\n",
		parm.pixels_per_line, parm.lines);
	break;
	
    default:
	if (parm.depth == 1)
	    fprintf(fout,"P4\n# SANE data follows\n%d %d\n",
		    parm.pixels_per_line, parm.lines);
	else
	    fprintf(fout,"P5\n# SANE data follows\n%d %d\n255\n",
		    parm.pixels_per_line, parm.lines);
	break;
    }

#endif
    scanprogress = progress_new("Scanning","Receiving data...",
				(CancelCallback)cancel_scan_callback, d);
    gtk_main_iteration();
    cnt = 0;
    total = parm.bytes_per_line*parm.lines;
    switch(parm.format) {
    case SANE_FRAME_GRAY:
    case SANE_FRAME_RGB:
	while(1) {
	    status = sane_read(scan_handle,d,parm.bytes_per_line,&len);
	    if(status==SANE_STATUS_GOOD) {
		cnt+=len;
#ifdef GIMP_PLUGIN
		d+=len;
#else
		fwrite (d, 1, len, fout);
#endif
		progress_update(scanprogress,(100*cnt)/total);
		gtk_main_iteration();
	    } else
		break;
	}
	break;
    default:	/* This client does not know how to handle other types of data */
	break;
    }
    progress_free(scanprogress);
    gtk_main_iteration();
#ifndef GIMP_PLUGIN
    fclose(fout);
    free(d);
#endif
    
    sane_cancel(scan_handle);
#ifdef GIMP_PLUGIN
    if (output) {
	gimp_display_image (output);
	gimp_update_image (output);
	gimp_free_image (output);
    }
    gimp_noexit_quit();
#endif
    sane_close(scan_handle);
    sane_exit();
    gtk_exit(0);
}

void select_device_callback (GtkWidget *widget,
			     gpointer   client_data,
			     gpointer   call_data)
{
    seldev = (long)client_data;
}

void button_callback (GtkWidget *widget,
		      gpointer   client_data,
		      gpointer   call_data)
{
    SANE_Word info;
    
    sane_control_option(scan_handle,(long)client_data,SANE_ACTION_SET_VALUE,NULL,&info);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
}

void info_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
}

void menu_callback (GtkWidget *widget,
		    gpointer   client_data,
		    gpointer   call_data)
{
    SANE_Word x, sel, info;

    x = (((long)client_data)&0xffff0000)>>16;
    sel = (((long)client_data)&0xffff);

    sane_control_option(scan_handle,x,SANE_ACTION_SET_VALUE,
			elements[x].menu[sel].label,&info);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
}

void draw_rect(GdkWindow *win, gint x1, gint y1, gint x2, gint y2)
{
    gint xx1, yy1, width, height;

    if(x1<=x2) {
	xx1 = x1;
	width = x2-x1;
    } else {
	xx1 = x2;
	width = x1-x2;
    }

    if(y1<=y2) {
	yy1 = y1;
	height = y2-y1;
    } else {
	yy1 = y2;
	height = y1-y2;
    }
    
    gdk_draw_rectangle(win,ScanWin.gc,FALSE,xx1,yy1,width,height);
}

void draw_selection(gint Redraw)
{
    gdk_gc_set_foreground(ScanWin.gc,&RGBPalette[1]);
    gdk_gc_set_background(ScanWin.gc,&RGBPalette[0]);
    gdk_gc_set_function(ScanWin.gc,GDK_INVERT);
    gdk_gc_set_line_attributes (ScanWin.gc,1,
				GDK_LINE_ON_OFF_DASH,
				GDK_CAP_BUTT,
				GDK_JOIN_MITER);
    if((Redraw)&&(OldSelection.active))
	draw_rect(ScanWin.preview->window,
		  OldSelection.x1, OldSelection.y1, 
		  OldSelection.x2, OldSelection.y2);

    if(Selection.active)
	draw_rect(ScanWin.preview->window,
		  Selection.x1, Selection.y1, 
		  Selection.x2, Selection.y2);
    OldSelection = Selection;
}

SANE_Word set_sane_option(gint setting, gint *value)
{
    SANE_Word info, len;
    const SANE_Option_Descriptor *opt;

    if(setting<0)
	return -1;

    opt=sane_get_option_descriptor(scan_handle,(long)setting);
    switch(opt->type) {
    case SANE_TYPE_INT:
	len = (long)*value;
	break;
    case SANE_TYPE_FIXED: 
	len = ((long)*value)<<SANE_FIXED_SCALE_SHIFT;
	break;
    default: 
	break;
    }
    sane_control_option(scan_handle,(long)setting,
			SANE_ACTION_SET_VALUE,&len,&info);
    if (info & SANE_INFO_INEXACT) {
	switch(opt->type) {
	case SANE_TYPE_INT:
	    break;
	case SANE_TYPE_FIXED: 
	    len>>=SANE_FIXED_SCALE_SHIFT;
	    break;
	default:
	    break;
	}
	*value = len;
    }
    return info;
}

void set_window(gint *x1, gint *y1, gint *x2, gint *y2)
{
    SANE_Word info;
    *x1 /= Preview.Ratio;
    *y1 /= Preview.Ratio;
    *x2 /= Preview.Ratio;
    *y2 /= Preview.Ratio;

    info = set_sane_option(DriverInfo.x1,x1);
    info |= set_sane_option(DriverInfo.y1,y1);
    info |= set_sane_option(DriverInfo.x2,x2);
    info |= set_sane_option(DriverInfo.y2,y2);
    update_complete_dialog();
    
    *x1 *= Preview.Ratio;
    *y1 *= Preview.Ratio;
    *x2 *= Preview.Ratio;
    *y2 *= Preview.Ratio;
    
    update_param();
}

void set_max_selection()
{
    const SANE_Option_Descriptor *opt;

    if(DriverInfo.x1>=0) {
	opt=sane_get_option_descriptor(scan_handle,DriverInfo.x1);
	Selection.x1 = (opt->constraint.range->min>>SANE_FIXED_SCALE_SHIFT)*Preview.Ratio;
    }
    if(DriverInfo.y1>=0) {
	opt=sane_get_option_descriptor(scan_handle,DriverInfo.y1);
	Selection.y1 = (opt->constraint.range->min>>SANE_FIXED_SCALE_SHIFT)*Preview.Ratio;
    }
    if(DriverInfo.x2>=0) {
	opt=sane_get_option_descriptor(scan_handle,DriverInfo.x2);
	Selection.x2 = (opt->constraint.range->max>>SANE_FIXED_SCALE_SHIFT)*Preview.Ratio;
    }
    if(DriverInfo.y2>=0) {
	opt=sane_get_option_descriptor(scan_handle,DriverInfo.y2);
	Selection.y2 = (opt->constraint.range->max>>SANE_FIXED_SCALE_SHIFT)*Preview.Ratio;
    }
    Selection.active = TRUE;
}

gint canvas_event (GtkWidget *canvas, GdkEvent *event)
{
    gint tmp;
    
    switch (event->type) {
    case GDK_EXPOSE:
	if(!ScanWin.gc) 
	    ScanWin.gc=gdk_gc_new(ScanWin.preview->window);
	draw_selection(FALSE);
        break; 
    case GDK_ENTER_NOTIFY:
	break;
    case GDK_LEAVE_NOTIFY:
        break;
    case GDK_BUTTON_PRESS:
	Selection.x1 = event->button.x;
	Selection.y1 = event->button.y;
	LeftButtonPressed = TRUE;
        break;
    case GDK_BUTTON_RELEASE:
	Selection.x2 = event->button.x;
	Selection.y2 = event->button.y;
	if((Selection.x1==Selection.x2)&&(Selection.y1==Selection.y2))
	    set_max_selection();
	else
	    Selection.active = TRUE;
	if(Selection.x1>Selection.x2) {
	    tmp = Selection.x1;
	    Selection.x1 = Selection.x2;
	    Selection.x2 = tmp;
	}
	if(Selection.y1>Selection.y2) {
	    tmp = Selection.y1;
	    Selection.y1 = Selection.y2;
	    Selection.y2 = tmp;
	}
                                
	set_window(&Selection.x1,&Selection.y1,
		   &Selection.x2,&Selection.y2);
	draw_selection(TRUE);
	LeftButtonPressed = FALSE;
        break;
    case GDK_MOTION_NOTIFY:
	if(LeftButtonPressed) {
	    Selection.x2 = event->motion.x;
	    Selection.y2 = event->motion.y;
	    if((Selection.x1==Selection.x2)&&(Selection.y1==Selection.y2))
		Selection.active = FALSE;
	    else
		Selection.active = TRUE;
	    draw_selection(TRUE);
	}
	ScanWin.hruledata->value = (event->motion.x/Preview.Ratio);
	ScanWin.vruledata->value = (event->motion.y/Preview.Ratio);
	gtk_data_notify ((GtkData *) ScanWin.hruledata);
	gtk_data_notify ((GtkData *) ScanWin.vruledata);
        break;
    default:
        break;
    }
    return(FALSE);
}

void dummy_callback (GtkWidget *widget,
		     gpointer   client_data,
		     gpointer   call_data)
{
    return;
}

static gint toggle_callback_update (GtkObserver *observer,
				    GtkData *data)
{
    GtkDataInt *intdata;
    gint setting;
    SANE_Word info;
    
    intdata = (GtkDataInt *) data;
    if((intdata->value==GTK_STATE_ACTIVATED)||(intdata->value==GTK_STATE_DEACTIVATED)) {
	switch(intdata->value) {
	case GTK_STATE_ACTIVATED:
	    setting = 1;
	    break;
	case GTK_STATE_DEACTIVATED:
	    setting = 0;
	    break;
	}
	sane_control_option(scan_handle,(long)observer->user_data,
			    SANE_ACTION_SET_VALUE,&setting,&info);
	if (info & SANE_INFO_INEXACT) {
	    gtk_check_button_set_state(elements[(long)observer->user_data].widget,
				       setting);
	}
	if (info & SANE_INFO_RELOAD_OPTIONS)
	    update_complete_dialog();
	if (info & SANE_INFO_RELOAD_PARAMS)
	    update_param();
    }

    return 0;
}

static gint scale_callback_update (GtkObserver *observer,
				   GtkData *data)
{
    GtkDataAdjustment *adjdata;
    SANE_Word info, len;
    const SANE_Option_Descriptor *opt;

    opt=sane_get_option_descriptor(scan_handle,(long)observer->user_data);
    
    adjdata = (GtkDataAdjustment *)data;

    switch(opt->type) {
    case SANE_TYPE_INT:
	len = (long)adjdata->value;
	break;
    case SANE_TYPE_FIXED: 
	len = ((long)adjdata->value)<<SANE_FIXED_SCALE_SHIFT;
	break;
    default: 
	break;
    }

    sane_control_option(scan_handle,(long)observer->user_data,
			SANE_ACTION_SET_VALUE,&len,&info);
    
    if (info & SANE_INFO_INEXACT) {
	switch(opt->type) {
	case SANE_TYPE_INT:
	    break;
	case SANE_TYPE_FIXED: 
	    len>>=SANE_FIXED_SCALE_SHIFT;
	    break;
	default:
	    break;
	}
	adjdata->value = (double) (len) * adjdata->step_increment;
	gtk_data_notify (data);
    }
    if (info & SANE_INFO_RELOAD_OPTIONS) 
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS) 
	update_param();
    return 0;
}

static void callback_disconnect (GtkObserver *observer,
				 GtkData *data)
{
}

static gint entry_callback (guint    keyval,
			    guint    state,
			    gpointer client_data)
{
    SANE_Word info;
    gchar buf[512];
    gchar *text;
    gint return_val;
    
    return_val = FALSE;
    text = gtk_text_entry_get_text (elements[(glong)client_data].widget);
    strcpy(buf,text);
    sane_control_option(scan_handle, (glong)client_data, SANE_ACTION_SET_VALUE, buf, &info);
    if (info & SANE_INFO_INEXACT)
	gtk_text_entry_set_text(elements[(glong)client_data].widget,buf);
    if (info & SANE_INFO_RELOAD_OPTIONS)
	update_complete_dialog();
    if (info & SANE_INFO_RELOAD_PARAMS)
	update_param();
    if (keyval == XK_Tab)
	return_val = TRUE;
    else if (keyval == XK_Return)
	return_val = TRUE;
    return(return_val);
}

/*
 * Find out whether a particular capability exists
 */
void match_option(gint *value, gint optnum, const gchar *dopt, gchar *opt)
{
    if(!dopt)
	return;
    if(!strcmp(dopt,opt))
	*value = optnum;
}

/*
 * Fetch the device capabilities: min/max coordinates and dpi
 */
void get_device_caps()
{
    gint x;
    const SANE_Option_Descriptor *opt;
    
    DriverInfo.dpi = -1;
    DriverInfo.x1 = -1;
    DriverInfo.y1 = -1;
    DriverInfo.x2 = -1;
    DriverInfo.y2 = -1;

    for(x=1;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++) {
	/* Find out if the options are going to be needed for the preview */
	match_option(&DriverInfo.dpi,x,opt->name,SANE_NAME_SCAN_RESOLUTION);
	match_option(&DriverInfo.x1,x,opt->name,SANE_NAME_SCAN_TL_X);
	match_option(&DriverInfo.y1,x,opt->name,SANE_NAME_SCAN_TL_Y);
	match_option(&DriverInfo.x2,x,opt->name,SANE_NAME_SCAN_BR_X);
	match_option(&DriverInfo.y2,x,opt->name,SANE_NAME_SCAN_BR_Y);
    }
}

/*
 * Set the scan sizes
 */

void get_scan_size()
{
    const SANE_Option_Descriptor *opt;
    
    if(DriverInfo.x2>=0) {
	opt = sane_get_option_descriptor(scan_handle,DriverInfo.x2);
	Preview.PhysX1 = SANE_UNFIX(opt->constraint.range->min);
	Preview.PhysX2 = SANE_UNFIX(opt->constraint.range->max);
	Preview.PhysWidth = Preview.PhysX2-Preview.PhysX1;
    } else {
	Preview.PhysX1 = 0;
	Preview.PhysX2 = 100;
	Preview.PhysWidth = 100;
    }
    if(DriverInfo.y2>=0) {
	opt = sane_get_option_descriptor(scan_handle,DriverInfo.y2);
	Preview.PhysY1 = SANE_UNFIX(opt->constraint.range->min);
	Preview.PhysY2 = SANE_UNFIX(opt->constraint.range->max);
	Preview.PhysHeight = Preview.PhysY2-Preview.PhysY1;
    } else {
	Preview.PhysY1 = 0;
	Preview.PhysY2 = 100;
	Preview.PhysHeight = 100;
    }
    Preview.Ratio = 500.0/(float)Preview.PhysHeight;
    Preview.PixelWidth = (long)(Preview.PhysWidth*Preview.Ratio);
    Preview.PixelHeight = (long)(Preview.PhysHeight*Preview.Ratio);
    Preview.PixelX1 = 0;
    Preview.PixelY1 = 0;
    Preview.PixelX2 = Preview.PixelWidth;
    Preview.PixelY2 = Preview.PixelHeight;
}

/*
 * Create the main dialog box
 */
static void device_dialog ()
{
    GtkWidget *main_vbox, *vbox, *hbox, *button, *label;
    GtkWidget *table, *frame, *parent;
    GtkWidget *entry, *hbox1, *scale, *menu, *menu_list;
    MenuItem *menu_items;
    gchar buf[1000];
    GtkDataAdjustment *adjustment;
    GtkObserver *observer;
    const gchar *devname;
    SANE_Status status;
    const SANE_Option_Descriptor *opt;
    SANE_Word sca;
    gint x, min, max, n, i;

    devname = devlist[seldev]->name;
    status = sane_open(devname,&scan_handle);
    if(status!=SANE_STATUS_GOOD) {
	error_dialog("Could not open selected device !");
	return;
    }

    get_device_caps();
    get_scan_size();

    set_max_selection();

    /* Create the dialog box */
    ScanWin.shell = gtk_window_new((char *) devname, GTK_WINDOW_TOPLEVEL);

    /* Create the main vbox */
    main_vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(main_vbox, 0);
    gtk_container_add(ScanWin.shell, main_vbox);
    gtk_widget_show(main_vbox);

    /* Top portion of window - preview and options */
    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(hbox, 3);
    gtk_box_pack(main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show(hbox);

    /* Column 1 - the preview */
    vbox = gtk_vbox_new(FALSE, 5);
    gtk_container_set_border_width(vbox, 3);
    gtk_box_pack(hbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show(vbox);

    /* The preview */
    table = gtk_table_new (2, 2, FALSE);
    gtk_table_set_col_spacing (table, 0, 1);
    gtk_table_set_row_spacing (table, 0, 1);
    gtk_container_set_border_width (table, 2);
    gtk_container_add (vbox, table);

    /* The empty box in the top-left corner */
    frame = gtk_frame_new(NULL);
    gtk_frame_set_type (frame, GTK_SHADOW_OUT);
    gtk_table_attach (table, frame, 0, 1, 0, 1, FALSE, FALSE, TRUE, 0,
		      FALSE, FALSE, TRUE, 0);
    gtk_widget_show (frame);

    /* Set the horizontal ruler */
    ScanWin.hruledata = (GtkDataAdjustment *) 
	gtk_data_adjustment_new (0, Preview.PhysX1, Preview.PhysX2, 
				 1, 1, 1);
    ScanWin.hruler = gtk_hruler_new (ScanWin.hruledata);
    gtk_ruler_set_metric (ScanWin.hruler, PIXELS);
    gtk_table_attach (table, ScanWin.hruler, 1, 2, 0, 1, TRUE, TRUE, TRUE, 0, FALSE,
		      FALSE, TRUE, 0);
    gtk_widget_show (ScanWin.hruler);

    /* Set the vertical ruler */
    ScanWin.vruledata = (GtkDataAdjustment *) 
	gtk_data_adjustment_new (0, Preview.PhysY1, Preview.PhysY2,
				 1, 1, 1);
    ScanWin.vruler = gtk_vruler_new (ScanWin.vruledata);
    gtk_ruler_set_metric (ScanWin.vruler, PIXELS);
    gtk_table_attach (table, ScanWin.vruler, 0, 1, 1, 2, FALSE, FALSE, TRUE, 0, 
		      TRUE, TRUE, TRUE, 0);
    gtk_widget_show (ScanWin.vruler);

    /* Set the preview area */
    frame = gtk_frame_new(NULL);
    gtk_frame_set_type(frame, GTK_SHADOW_IN);
    ScanWin.gc = NULL;
    ScanWin.preview = 
	gtk_drawing_area_new (Preview.PixelWidth, Preview.PixelHeight, canvas_event,
			      GDK_EXPOSURE_MASK | GDK_ENTER_NOTIFY_MASK | 
			      GDK_LEAVE_NOTIFY_MASK | GDK_POINTER_MOTION_MASK | 
			      GDK_BUTTON_PRESS_MASK | GDK_BUTTON_RELEASE_MASK);
    gtk_container_add(frame, ScanWin.preview);
    gtk_widget_show(ScanWin.preview);
    gtk_widget_show(frame);
    Selection.active = TRUE;
    OldSelection.active = FALSE;
    LeftButtonPressed = FALSE;

    gtk_table_attach (table, frame, 1, 2, 1, 2, TRUE, TRUE, TRUE, 0, TRUE,
		      TRUE, TRUE, 0);
    gtk_widget_show (table);


    /* Column 2 - the options */
    vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_set_border_width (vbox, 3);
    gtk_box_pack (hbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (vbox);
    
    parent = vbox;
    /* Build the options from the backend info */
    for(x=1;(opt=sane_get_option_descriptor(scan_handle,x))!=NULL;x++) {
	switch(opt->type) { 
	case SANE_TYPE_GROUP: /* Group a set of options */
	    frame = gtk_frame_new ((char *) opt->title);
	    gtk_container_set_border_width(frame, 4);
	    gtk_frame_set_type(frame, GTK_SHADOW_ETCHED_IN);
	    gtk_box_pack(vbox, frame, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_widget_show (frame);
	    
	    parent = gtk_vbox_new(FALSE, 4);
	    gtk_container_set_border_width(parent, 2);
	    gtk_container_add(frame, parent);
	    gtk_widget_show(parent);
	    elements[x].widget = parent;
	    elements[x].label = NULL;
	    break;
	case SANE_TYPE_BOOL: /* We use a checkbutton for bool */
	    if (opt->size != sizeof(SANE_Word)) {
		fprintf(stderr, "Boolean vectors not supported yet\n");
		break;
	    }

	    button = gtk_check_button_new ();
	    observer = (GtkObserver *)malloc(sizeof(GtkObserver));
	    observer->update = toggle_callback_update;
	    observer->disconnect = callback_disconnect;
	    observer->user_data = (void *)(long) x;
	    gtk_data_attach(gtk_button_get_state(button),observer);
	    
	    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
	    gtk_check_button_set_state(button,sca);

	    label = gtk_label_new ((char *) opt->title);
	    gtk_label_set_alignment (label, 0.0, 0.5);
	    gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_container_add (button, label);
	    gtk_widget_show (label);
	    gtk_widget_show (button);
	    elements[x].widget = button;
	    elements[x].label = label;
	    break;
	case SANE_TYPE_INT:
	    {
		if (opt->size != sizeof(SANE_Word)) {
		    fprintf(stderr, "Integer vectors not supported yet\n");
		    break;
		}
		switch(opt->constraint_type) { 
		case SANE_CONSTRAINT_RANGE: /* Use a scale */
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new ((char *) opt->title);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    min = opt->constraint.range->min;
		    max = opt->constraint.range->max;
		    adjustment = (GtkDataAdjustment*)
			gtk_data_adjustment_new (sca, min, max, 1, 1, 10);
		    scale = gtk_hscale_new (adjustment);
		    gtk_scale_set_value_pos (scale, GTK_POS_TOP);
		    gtk_scale_set_digits (scale, 0);
		    gtk_box_pack (hbox1, scale, TRUE, TRUE, 0, GTK_PACK_START);
		    gtk_widget_show (label);
		    gtk_widget_show (scale);
		    gtk_widget_show (hbox1);
		    elements[x].widget = scale;
		    elements[x].label = label;
		    elements[x].data = (GtkData *)adjustment;
		    break;
		case SANE_CONSTRAINT_WORD_LIST: /* Use a "list-select" widget */
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new ((char *) opt->title);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    menu = gtk_option_menu_new ();
		    gtk_box_pack (hbox1, menu, FALSE, FALSE, 2, GTK_PACK_START);
		    n = opt->constraint.word_list[0];
		    menu_items = (MenuItem *)malloc((n+1)*sizeof(MenuItem));
		    for(i=0;i<n;i++) {
			sprintf(buf,"%d",opt->constraint.word_list[i+1]);
			menu_items[i].label = strdup(buf);
			menu_items[i].accelerator_key = 0;
			menu_items[i].accelerator_mods = 0;
			menu_items[i].callback = menu_callback;
			menu_items[i].user_data = (gpointer)(long)(i+(x<<16));
			menu_items[i].subitems = NULL;
			menu_items[i].widget = NULL;
		    }
		    menu_items[n].label = NULL;
		    menu_items[n].accelerator_key = 0;
		    menu_items[n].accelerator_mods = 0;
		    menu_items[n].callback = NULL;
		    menu_items[n].user_data = NULL;
		    menu_items[n].subitems = NULL;
		    menu_items[n].widget = NULL;
		    /* Fill in menu_items */
		    menu_list = build_menu(menu_items,NULL);
		    gtk_widget_show (label);
		    gtk_widget_show (menu);
		    gtk_widget_show (hbox1);
		    gtk_option_menu_set_menu (menu, menu_list);
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		    sprintf(buf,"%d",sca);
		    gtk_option_menu_set_history(menu,find_menu_option(menu_items,buf));
		    elements[x].widget = menu;
		    elements[x].label = label;
		    elements[x].menu = menu_items;
		    break;
		default:
		    break;
		}
		break;
	    }
	case SANE_TYPE_FIXED: 
	    {
		if (opt->size != sizeof(SANE_Word)) {
		    fprintf(stderr, "Fixed vectors not supported yet\n");
		    break;
		}
		switch(opt->constraint_type) { 
		case SANE_CONSTRAINT_RANGE: /* Use a scale */
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new ((char *) opt->title);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    min = opt->constraint.range->min>>SANE_FIXED_SCALE_SHIFT;
		    max = opt->constraint.range->max>>SANE_FIXED_SCALE_SHIFT;
		    sca = sca>>SANE_FIXED_SCALE_SHIFT;
		    adjustment = (GtkDataAdjustment*) 
			gtk_data_adjustment_new (sca, min, max, 1, 1, 10);
		    
		    scale = gtk_hscale_new (adjustment);
		    gtk_scale_set_value_pos (scale, GTK_POS_TOP);
		    gtk_scale_set_digits (scale, 0);

		    observer = (GtkObserver *)malloc(sizeof(GtkObserver));
		    observer->update = scale_callback_update;
		    observer->disconnect = callback_disconnect;
		    observer->user_data = (void *)(long) x;
		    gtk_data_attach ((GtkData *)adjustment, observer);

		    gtk_box_pack (hbox1, scale, TRUE, TRUE, 0, GTK_PACK_START);
		    gtk_widget_show (label);
		    gtk_widget_show (scale);
		    gtk_widget_show (hbox1);
		    elements[x].widget = scale;
		    elements[x].label = label;
		    elements[x].data = (GtkData *)adjustment;
		    break;
		case SANE_CONSTRAINT_WORD_LIST: /* Use a "list-select" widget */
		    hbox1 = gtk_hbox_new (FALSE, 2);
		    gtk_container_set_border_width (hbox1, 0);
		    gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		    label = gtk_label_new ((char *) opt->title);
		    gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		    menu = gtk_option_menu_new ();
		    gtk_box_pack (hbox1, menu, FALSE, FALSE, 2, GTK_PACK_START);
		    n = opt->constraint.word_list[0];
		    menu_items = (MenuItem *)malloc((n+1)*sizeof(MenuItem));
		    for(i=0;i<n;i++) {
			sprintf(buf,"%d",opt->constraint.word_list[i+1]);
			menu_items[i].label = strdup(buf);
			menu_items[i].accelerator_key = 0;
			menu_items[i].accelerator_mods = 0;
			menu_items[i].callback = menu_callback;
			menu_items[i].user_data = (gpointer)(long)(i+(x<<16));
			menu_items[i].subitems = NULL;
			menu_items[i].widget = NULL;
		    }
		    menu_items[n].label = NULL;
		    menu_items[n].accelerator_key = 0;
		    menu_items[n].accelerator_mods = 0;
		    menu_items[n].callback = NULL;
		    menu_items[n].user_data = NULL;
		    menu_items[n].subitems = NULL;
		    menu_items[n].widget = NULL;
		    /* Fill in menu_items */
		    menu_list = build_menu(menu_items,NULL);
		    gtk_widget_show (label);
		    gtk_widget_show (menu);
		    gtk_widget_show (hbox1);
		    gtk_option_menu_set_menu (menu, menu_list);
		    sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,&sca,NULL);
		    sprintf(buf,"%d",sca);
		    gtk_option_menu_set_history(menu,find_menu_option(menu_items,buf));
		    elements[x].widget = menu;
		    elements[x].label = label;
		    elements[x].menu = menu_items;
		    break;
		default:
		    break;
		}
		break;
	    } 
	case SANE_TYPE_STRING:
	    switch(opt->constraint_type) {
	    case SANE_CONSTRAINT_STRING_LIST:
		hbox1 = gtk_hbox_new (FALSE, 2);
		gtk_container_set_border_width (hbox1, 0);
		gtk_box_pack (parent, hbox1, FALSE, FALSE, 0, GTK_PACK_START);
		label = gtk_label_new ((char *) opt->title);
		gtk_box_pack (hbox1, label, FALSE, FALSE, 2, GTK_PACK_START);
		menu = gtk_option_menu_new ();
		gtk_box_pack (hbox1, menu, FALSE, FALSE, 2, GTK_PACK_START);
		n = 0;
		while(opt->constraint.string_list[n])
		    ++n;
		menu_items = (MenuItem *)malloc((n+1)*sizeof(MenuItem));
		for(i=0;i<n;i++) {
		    menu_items[i].label = (char *) opt->constraint.string_list[i];
		    menu_items[i].accelerator_key = 0;
		    menu_items[i].accelerator_mods = 0;
		    menu_items[i].callback = menu_callback;
		    menu_items[i].user_data = (gpointer)(long)(i+(x<<16));
		    menu_items[i].subitems = NULL;
		    menu_items[i].widget = NULL;
		}
		menu_items[n].label = NULL;
		menu_items[n].accelerator_key = 0;
		menu_items[n].accelerator_mods = 0;
		menu_items[n].callback = NULL;
		menu_items[n].user_data = NULL;
		menu_items[n].subitems = NULL;
		menu_items[n].widget = NULL;
		/* Fill in menu_items */
		menu_list = build_menu(menu_items,NULL);
		gtk_widget_show (label);
		gtk_widget_show (menu);
		gtk_widget_show (hbox1);
		gtk_option_menu_set_menu (menu, menu_list);
		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
		gtk_option_menu_set_history(menu,find_menu_option(menu_items,buf));
		elements[x].widget = menu;
		elements[x].label = label;
		elements[x].menu = menu_items;
		break;
	    case SANE_CONSTRAINT_NONE:
		entry = gtk_text_entry_new ();
		gtk_text_entry_set_key_function (entry, entry_callback, (void *)(long)x);
		sane_control_option(scan_handle,x,SANE_ACTION_GET_VALUE,buf,NULL);
		gtk_text_entry_set_text (entry, buf);
		gtk_box_pack (parent, entry, FALSE, TRUE, 0, GTK_PACK_START);
		gtk_widget_show (entry);
		elements[x].widget = entry;
		elements[x].label = NULL;
		break;
	    default:
		break;
	    }
	    break;
	case SANE_TYPE_BUTTON:
	    button = gtk_push_button_new ();
	    gtk_callback_add (gtk_button_get_state (button), button_callback, (void *)(long)x);
	    label = gtk_label_new ((char *) opt->title);
	    gtk_box_pack (parent, button, FALSE, TRUE, 0, GTK_PACK_START);
	    gtk_container_add (button, label);
	    gtk_widget_show (label);
	    gtk_widget_show (button);
	    elements[x].widget = button;
	    elements[x].label = label;
	    break;
	default:
	    break;
	}
    }

    /* The info row */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);

    frame = gtk_frame_new(NULL);
    gtk_frame_set_type(frame, GTK_SHADOW_IN);
    gtk_box_pack(hbox, frame, FALSE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show(frame);

    hbox = gtk_hbox_new(FALSE, 5);
    gtk_container_set_border_width(hbox, 2);
    gtk_container_add(frame, hbox);
    gtk_widget_show(hbox);

    ScanWin.info_label = gtk_label_new ("0 x 0 : 0 KB");
    gtk_label_set_alignment (ScanWin.info_label, 0.0, 0.5);
    gtk_box_pack (hbox, ScanWin.info_label, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (ScanWin.info_label);
    
    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);
    
    /* The Scan button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), scan_dialog, NULL);
    label = gtk_label_new ("Scan");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    /* The Preview button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), scan_preview, NULL);
    label = gtk_label_new ("Preview");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    /* The Zoom in button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), zoom_in_preview, NULL);
    label = gtk_label_new ("Zoom in");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);

    /* The Zoom out button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), zoom_out_preview, NULL);
    label = gtk_label_new ("Zoom out");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    /* The Cancel button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), quit_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    gtk_widget_show(ScanWin.shell);
}

void ok_choose_dialog_callback (GtkWidget *widget,
				gpointer   client_data,
				gpointer   call_data)
{
    gtk_widget_hide(dialog);
    gtk_widget_destroy(dialog);
    device_dialog ();
}

static void choose_device ()
{
    GtkWidget *main_vbox, *vbox, *hbox, *button, *label;
    GtkData *owner;
    const SANE_Device *adev;
    gint i;

   
    dialog = gtk_window_new("Select device", GTK_WINDOW_TOPLEVEL);
    
    main_vbox = gtk_vbox_new(FALSE,5);
    gtk_container_set_border_width (main_vbox, 0);
    gtk_container_add (dialog, main_vbox);
    gtk_widget_show (main_vbox);
    
    /* The list of drivers */
    vbox = gtk_vbox_new (FALSE, 5);
    gtk_container_set_border_width (vbox, 3);
    gtk_box_pack (main_vbox, vbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (vbox);
    
    /* The radio buttons */
    owner = NULL;
    for(i=0;i<ndevs;i++) {
	adev = devlist[i];
	
	button = gtk_radio_button_new (owner);
	gtk_callback_add(gtk_button_get_state(button),
			 select_device_callback,
			 (void *)(long) i);
	label = gtk_label_new ((char *) adev->name);
	gtk_label_set_alignment (label, 0.0, 0.5);
	gtk_box_pack (vbox, button, TRUE, TRUE, 0, GTK_PACK_START);
	gtk_container_add (button, label);
	gtk_widget_show (label);
	gtk_widget_show (button);
	if(i==0)
	    owner = gtk_button_get_owner (button);
    }
    
    /* The bottom row of buttons */
    hbox = gtk_hbox_new (FALSE, 5);
    gtk_container_set_border_width (hbox, 3);
    gtk_box_pack (main_vbox, hbox, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_widget_show (hbox);
    
    /* The Ok button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), ok_choose_dialog_callback, NULL);
    label = gtk_label_new ("Ok");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    /* The Cancel button */
    button = gtk_push_button_new ();
    gtk_callback_add (gtk_button_get_state (button), quit_callback, NULL);
    label = gtk_label_new ("Cancel");
    gtk_box_pack (hbox, button, TRUE, TRUE, 0, GTK_PACK_START);
    gtk_container_add (button, label);
    gtk_widget_show (label);
    gtk_widget_show (button);
    
    gtk_widget_show (dialog);
}

int main (int argc, char **argv)
{
    const SANE_Device *adev;
    ScanWin.info_label=NULL;
    scan_handle=NULL;
#ifdef GIMP_PLUGIN
    output = NULL;
    if(gimp_init (argc, argv)) {
#endif
	gdk_set_debug_level (0);
	gdk_set_show_events (0);
	gtk_init (&argc, &argv);
	
	sane_init(NULL);
	
	sane_get_devices(&devlist, SANE_FALSE);
	
	ndevs = 0;
	adev = devlist[0];
	while(adev)
	    adev = devlist[++ndevs];
	if(ndevs>0) {
	    
	    if(ndevs==1) {
		seldev = 0;
		device_dialog();
	    } else
		choose_device();

	    gtk_main();
	} else
	    fprintf(stderr,"No devices available\n");
	
	sane_exit();
#ifdef GIMP_PLUGIN
	gimp_noexit_quit();
#endif
	gtk_exit(0);
#ifdef GIMP_PLUGIN    
    } else 
	fprintf(stderr,"This module can only be invoked from within the GIMP\n");
#endif
    return 0;
}
